/*
 * argument.c
 *
 * Copyright 1998, 1999 Michael Elizabeth Chastain, <mailto:mec@shout.net>.
 * Licensed under the Gnu Public License, version 2.
 */

#include <sys/stat.h>
#include <sys/utsname.h>

#include <errno.h>
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "mconfig.h"

static const char str_version [] = "0.17";



/*
 * Command arguments.
 */

enum
{
    opt_help = 1000,
    opt_version,
    opt_mode,
    opt_arch,
    opt_seed,
    opt_title,
    opt_ds,			/* $dir-source			*/
    opt_dt,			/* $dir-target			*/
    opt_fs,			/* $file-script			*/
    opt_fci, opt_no_fci,	/* $file-config-in		*/
    opt_fco, opt_no_fco,	/* $file-config-out		*/
    opt_fac, opt_no_fac,	/* $file-autoconf-h		*/
    opt_no_asm_symlink,		/* no include/asm symlink	*/
    opt_no_backup,		/* no .config.old file		*/
};

static const struct option options [] =
{
    { "help",			0, NULL, opt_help		},
    { "version",		0, NULL, opt_version		},
    { "mode",			1, NULL, opt_mode		},
    { "architecture",		1, NULL, opt_arch		},
    { "seed",			1, NULL, opt_seed		},
    { "title",			1, NULL, opt_title		},
    { "dir-source",		1, NULL, opt_ds			},
    { "ds",			1, NULL, opt_ds			},
    { "dir-target",		1, NULL, opt_dt			},
    { "dt",			1, NULL, opt_dt			},
    { "file-script",		1, NULL, opt_fs			},
    { "fs",			1, NULL, opt_fs			},
    { "file-config-in",		1, NULL, opt_fci		},
    { "fci",			1, NULL, opt_fci		},
    { "no-file-config-in",	0, NULL, opt_no_fci		},
    { "no-fci",			0, NULL, opt_no_fci		},
    { "file-config-out",	1, NULL, opt_fco		},
    { "fco",			1, NULL, opt_fco		},
    { "no-file-config-out",	0, NULL, opt_no_fco		},
    { "no-fco",			0, NULL, opt_no_fco		},
    { "file-autoconf-h",	1, NULL, opt_fac		},
    { "fac",			1, NULL, opt_fac		},
    { "no-file-autoconf-h",	0, NULL, opt_no_fac		},
    { "no-fac",			0, NULL, opt_no_fac		},
    { "no-asm-link",		0, NULL, opt_no_asm_symlink	},
    { "no-backup",		0, NULL, opt_no_backup		},
    { NULL,			0, NULL, 0			},
};

argument_type argument =
{
    str_version, NULL, NULL,
    mode_NULL, NULL, 0, NULL, NULL, NULL,
    NULL, NULL, NULL, NULL,
    0, 0
};



/*
 * Make a nice default title.
 * Read ${dir-source}Makefile to find the version.
 * FIXME: what should I do about white space in input files?
 */

static const char * title_default( const char * dir_source, const char * arch )
{
    char version    [64] = "";
    char patchlevel [64] = "";
    char sublevel   [64] = "";
    char extra      [64] = "";
    char name       [64] = "";
    const char * title;
    char * filename;
    FILE * fp;
    char * buf;

    title = "Linux Kernel Configuration (unknown version)";

    filename = grab_memory( strlen(dir_source) + 16 + 1 );
    sprintf( filename, "%s%s", dir_source, "Makefile" );
    fp = fopen( filename, "r" );
    if ( fp == NULL )
	return title;

    if ( fscanf( fp, "VERSION = %63s\n",    version    ) == 1
    &&   fscanf( fp, "PATCHLEVEL = %63s\n", patchlevel ) == 1
    &&   fscanf( fp, "SUBLEVEL = %63s\n",   sublevel   ) == 1 )
    {
	/* EXTRAVERSION is optional */
	if ( fscanf( fp, "EXTRAVERSION =%*[ ]%63s", extra ) != 1 )
	    extra [0] = '\0';
    }

    if ( fclose( fp ) != 0 )
	error_system_call( filename );

    sprintf( filename, "%s%s", dir_source, ".name" );
    fp = fopen( filename, "r" );
    if ( fp != NULL )
    {
	if ( fscanf( fp, "%63s\n", name ) != 1 )
	    name[0] = '\0';
	if ( fclose( fp ) != 0 )
	    error_system_call( filename );
    }

    buf = grab_memory( 48 + strlen(arch) + strlen(version) + strlen(patchlevel)
	+ strlen(sublevel) + strlen(extra) + strlen(name) );
    sprintf( buf, "Linux/%s v%s.%s.%s%s %s%s%sKernel Configuration",
	arch, version, patchlevel, sublevel, extra,
	name[0] ? "(" : "", name, name[0] ? ") " : "" );
    title = buf;

    return title;
}



/*
 * Parse arguments.
 * FIXME: also parse $MCONFIG from the environment.
 */

void argument_parse( int argc, char * argv [] )
{
    const char * seed	= NULL;
    int no_fci		= 0;
    int no_fco		= 0;
    int no_fac		= 0;

    /*
     * Set the command name.  This is first so that error reports come
     * out nice.  And yes, I am paranoid here, I want a nice message if
     * the very first grab_string fails; and I want to call grab_string
     * because getopt_long_only may modify the argv block later.
     */

    argument.cmdname = argv [0];
    argument.cmdname = grab_string( argv [0], strlen(argv[0]) );



    /*
     * Save original command for output.
     */

    {
	char * cmdline;
	int iarg;
	int len = 1;

	for ( iarg = 0; iarg < argc; ++iarg )
	    len += 1 + strlen(argv[iarg]);

	cmdline     = grab_memory( len );
	cmdline [0] = '\0';

	for ( iarg = 0; iarg < argc; ++iarg )
	{
	    if ( iarg > 0 )
		strcat( cmdline, " " );
	    strcat( cmdline, argv [iarg] );
	}

	argument.cmdline = cmdline;
    }



    /*
     * Process options.
     */

    for ( ; ; )
    {
	int c = getopt_long_only( argc, argv, "hvm:a:", options, NULL );
	if ( c == EOF )
	{
	    if ( optind < argc )
		error_exit( "extra arguments not allowed" );
	    break;
	}

	switch ( c )
	{
	default:
	    error_internal( "argument_parse: getopt_long failed" );
	    break;

	case ':':
	case '?':
	    exit( 1 );

	case 'h':
	case opt_help:
	    printf( "Usage: %s: options ...\n", argv [0]		     );
	    printf( "  --help\n"					     );
	    printf( "  --version\n"					     );
	    printf( "  -m MODE,     --mode MODE\n"			     );
	    printf( "     modes: menu x line old\n"			     );
	    printf( "     modes: syntax random mutate minimum maximum\n"     );
	    printf( "  -a ARCH,     --architecture ARCH\n"		     );
	    printf( "  --seed SEED\n"					     );
	    printf( "  --title TITLE\n"					     );
	    printf( "  --ds DIR,    --dir-source DIR\n"			     );
	    printf( "  --dt DIR,    --dir-target DIR\n"			     );
	    printf( "  --fs FILE,   --file-script FILE\n"		     );
	    printf( "  --fci FILE,  --file-config-in FILE\n"		     );
	    printf( "  --fco FILE,  --file-config-out FILE\n"		     );
	    printf( "  --fac FILE,  --file-autoconf-h FILE\n"		     );
	    printf( "  --no-fci, --no-fco, --no-fac\n"			     );
	    printf( "  --no-asm-symlink\n"				     );
	    printf( "  --no-backup\n"					     );
	    exit( 0 );

	case 'v':
	case opt_version:
	    printf( "%s version %s\n", argv [0], str_version );
	    exit( 0 );

	case 'm':
	case opt_mode:
	    if ( optarg == NULL || optarg [0] == '\0' )
		error_exit( "mode name required" );
	    if ( argument.mode != mode_NULL )
		error_exit( "mode already set" );
	    else if ( strncmp( optarg, "menu",    strlen(optarg) ) == 0 )
		argument.mode = mode_menu;
	    else if ( strncmp( optarg, "x-gtk",   strlen(optarg) ) == 0 )
		argument.mode = mode_x_gtk;
	    if      ( strncmp( optarg, "line",    strlen(optarg) ) == 0 )
		argument.mode = mode_line;
	    else if ( strncmp( optarg, "old",     strlen(optarg) ) == 0 )
		argument.mode = mode_old;
	    else if ( strncmp( optarg, "syntax",  strlen(optarg) ) == 0 )
		argument.mode = mode_syntax;
	    else if ( strncmp( optarg, "random",  strlen(optarg) ) == 0 )
		argument.mode = mode_random;
	    else if ( strncmp( optarg, "mutate",  strlen(optarg) ) == 0 )
		argument.mode = mode_mutate;
	    else if ( strncmp( optarg, "minimum", strlen(optarg) ) == 0 )
		argument.mode = mode_minimum;
	    else if ( strncmp( optarg, "maximum", strlen(optarg) ) == 0 )
		argument.mode = mode_maximum;
	    else
		error_exit( "unknown mode" );
	    break;

	case 'a':
	case opt_arch:
	    if ( optarg == NULL || optarg [0] == '\0' )
		error_exit( "architecture name required" );
	    if ( argument.arch != NULL )
		error_exit( "architecture already set" );
	    argument.arch = grab_string( optarg, strlen(optarg) );
	    break;

	case opt_seed:
	    {
		long int lseed;
		char * endptr;

		if ( optarg == NULL || optarg [0] == '\0' )
		    error_exit( "seed required" );
		if ( seed != NULL )
		    error_exit( "seed already specified" );
		seed = grab_string( optarg, strlen(optarg) );
		argument.useed = lseed = strtol( seed, & endptr, 0 );
		if ( argument.useed != lseed || * endptr != '\0' )
		    error_exit( "invalid seed" );
	    }
	    break;

	case opt_title:
	    if ( optarg == NULL || optarg [0] == '\0' )
		error_exit( "title required" );
	    if ( argument.title != NULL )
		error_exit( "title already set" );
	    argument.title = grab_string( optarg, strlen(optarg) );
	    break;

	case opt_ds:
	    if ( optarg == NULL || optarg [0] == '\0' )
		error_exit( "directory name required" );
	    if ( argument.ds != NULL )
		error_exit( "dir-source already set" );
	    {
		int len    = strlen(optarg);
		char * ptr = grab_memory( len+2 );
		memcpy( ptr, optarg, len+1 );
		if ( len > 0 && ptr [len] != '/' )
		{
		    ptr [len]   = '/';
		    ptr [len+1] = '\0';
		}
		argument.ds = ptr;
	    }
	    break;

	case opt_dt:
	    if ( optarg == NULL || optarg [0] == '\0' )
		error_exit( "directory name required" );
	    if ( argument.dt != NULL )
		error_exit( "dir-target already set" );
	    {
		int len = strlen(optarg);
		char * ptr = grab_memory( len+2 );
		memcpy( ptr, optarg, len+1 );
		if ( len > 0 && ptr [len] != '/' )
		{
		    ptr [len]   = '/';
		    ptr [len+1] = '\0';
		}
		argument.dt = ptr;
	    }
	    break;

	case opt_fs:
	    if ( optarg == NULL || optarg [0] == '\0' )
		error_exit( "file name required" );
	    if ( argument.fs != NULL )
		error_exit( "file-script already set" );
	    argument.fs = grab_string( optarg, strlen(optarg) );
	    break;
	    
	case opt_fci:
	    if ( optarg == NULL || optarg [0] == '\0' )
		error_exit( "file name required" );
	    if ( argument.fci != NULL )
		error_exit( "file-config-in already set" );
	    if ( no_fci )
		error_exit( "no-file-config-in already set" );
	    argument.fci = grab_string( optarg, strlen(optarg) );
	    break;

	case opt_no_fci:
	    if ( no_fci )
		error_exit( "no-file-config-in already set" );
	    if ( argument.fci )
		error_exit( "file-config-in already set" );
	    no_fci = 1;
	    break;

	case opt_fco:
	    if ( optarg == NULL || optarg [0] == '\0' )
		error_exit( "file name required" );
	    if ( argument.fco != NULL )
		error_exit( "file-config-out already set" );
	    if ( no_fco )
		error_exit( "no-file-config-out already set" );
	    argument.fco = grab_string( optarg, strlen(optarg) );
	    break;

	case opt_no_fco:
	    if ( no_fco )
		error_exit( "no-file-config-out already set" );
	    if ( argument.fco )
		error_exit( "file-config-out already set" );
	    no_fco = 1;
	    break;

	case opt_fac:
	    if ( optarg == NULL || optarg [0] == '\0' )
		error_exit( "file name required" );
	    if ( argument.fac != NULL )
		error_exit( "file-autoconf-h already set" );
	    if ( no_fac )
		error_exit( "no-file-autoconf-h already set" );
	    argument.fac = grab_string( optarg, strlen(optarg) );
	    break;

	case opt_no_fac:
	    if ( no_fac )
		error_exit( "no-file-autoconf-h already set" );
	    if ( argument.fac )
		error_exit( "file-autoconf-h already set" );
	    no_fac = 1;
	    break;

	case opt_no_asm_symlink:
	    if ( argument.no_asm_symlink )
		error_exit( "no-asm-link already set" );
	    argument.no_asm_symlink = 1;
	    break;

	case opt_no_backup:
	    if ( argument.no_backup )
		error_exit( "no-backup already set" );
	    argument.no_backup = 1;
	    break;
	}
    }



    /*
     * Set defaults.
     */

    /* mode defaults to --menu */
    if ( argument.mode == mode_NULL )
	argument.mode = mode_menu;

    /* arch defaults to host architecture */
    if ( argument.arch == NULL )
    {
	struct utsname buf;
	if ( uname( &buf ) != 0 )
	    error_system_call( "uname" );

	if      ( strcmp( buf.machine, "alpha" ) == 0 )
	    argument.arch = "alpha";

	else if ( buf.machine[0] == 'a' && buf.machine[1] == 'r'
	&&        buf.machine[2] == 'm' )
	    argument.arch = "arm";

	else if ( buf.machine[0] == 'i' && buf.machine[1] != '\0'
	&&        buf.machine[2] == '8' && buf.machine[3] == '6'
	&&        buf.machine[4] == '\0' )
	    argument.arch = "i386";

	else if ( strcmp( buf.machine, "m68k" ) == 0 )
	    argument.arch = "m68k";

	else if ( strcmp( buf.machine, "mips" ) == 0 )
	    argument.arch = "mips";

	else if ( strcmp( buf.machine, "ppc" ) == 0 )
	    argument.arch = "ppc";

	else if ( strcmp( buf.machine, "sparc" ) == 0 )
	    argument.arch = "sparc";

	else if ( strcmp( buf.machine, "sparc64" ) == 0 )
	    argument.arch = "sparc64";

	else
	    error_exit( "no arch selected and no known default" );
    }

    /* ds defaults to current directory */
    if ( argument.ds == NULL )
	argument.ds = "";

    /* dt defaults to $ds */
    if ( argument.dt == NULL )
	argument.dt = argument.ds;

    /* title defaults to a nice title */
    if ( argument.title == NULL )
	argument.title = title_default( argument.ds, argument.arch );

    /* seed defaults to something random */
    /* FIXME: use /dev/random, /dev/urandom */
    if ( seed == NULL )
	argument.useed = time( NULL ) + getpid( );

    /* fs defaults to $ds/arch/$arch/config.in */
    if ( argument.fs == NULL )
    {
	char * name = grab_memory( strlen(argument.ds) + strlen(argument.arch) + 32 );
	sprintf( name, "%sarch/%s/config.in", argument.ds, argument.arch );
	argument.fs = name;
    }

    /*
     * fci defaults to $dt/.config if it exists followed by
     * $ds/$ARCH/defconfig if it exists.
     */
    if ( ! no_fci && argument.fci == NULL )
    {
	struct stat stat_buf;
	char * name1 = grab_memory( strlen(argument.dt) + 32 );

	sprintf( name1, "%s.config", argument.dt );
	if ( stat( name1, &stat_buf ) == 0 || errno != ENOENT )
	{
	    argument.fci = name1;
	}
	else
	{
	    char * name2 =
		grab_memory( strlen(argument.ds) + strlen(argument.arch) + 32 );
	    sprintf( name2, "%sarch/%s/defconfig", argument.ds, argument.arch );
	    if ( stat( name2, &stat_buf ) == 0 || errno != ENOENT )
	    argument.fci = name2;
	}
    }

    /* fco defaults to $dt/.config */
    if ( ! no_fco && argument.fco == NULL )
    {
	char * name = grab_memory( strlen(argument.dt) + 32 );
	sprintf( name, "%s.config", argument.dt );
	argument.fco = name;
    }

    /* fac defaults to $dt/include/linux/autoconf.h */
    if ( ! no_fac && argument.fac == NULL )
    {
	char * name = grab_memory( strlen(argument.dt) + 32 );
	sprintf( name, "%sinclude/linux/autoconf.h", argument.dt );
	argument.fac = name;
    }

    /* no-asm-symlink defaults to off; no action needed */

    /* turn off some options in some non-interactive modes */
    switch ( argument.mode )
    {
    default:
	break;

    case mode_random:
    case mode_minimum:
    case mode_maximum:
	argument.fci       = NULL;
	argument.no_backup = 1;
	break;
    }
}
